/*
 * Copyright (c) 2004, 2005 Marcus Glocker <marcus@nazgul.ch>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <dirent.h>
#include <fcntl.h>
#include <openssl/ssl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "../libmy/str.h"
#include "../libmy/file.h"
#ifdef __linux__
#include <crypt.h>
#include "../libbsd/strlcpy.h"
#endif
#include "proto.h"
#include "extern.h"

/*
 * sys_mime()
 * 	searches mime type in our memory mapped mime list
 * Return:
 *	0 = mime not found, 1 = mime found
 */
int
sys_mime(char *dst, const int dsize, const char *mimes, const int msize,
    const char *type)
{
	int	i, j, k;
	char	*x, l[2][1024];
	
	for (i = 0, j = 0, k = 0; i <= msize; i++) {
		/* no control characters found, read */
		if (mimes[i] != ' ' && mimes[i] != '\t' && mimes[i] != '\n') {
			l[j][k] = mimes[i];
			k++;
		}
		/* second field starts */
		if (mimes[i] == ' ' || mimes[i] == '\t') {
			l[j][k] = '\0';
			j = 1;
			k = 0;
		}
		/* end of line */
		if (mimes[i] == '\n') {
			l[j][k] = '\0';
			for (x = strtok(l[1], ","); x; x = strtok(NULL, ",")) {
				if (strcasecmp(x, type) == 0) {
					strlcpy(dst, l[0], dsize);
					return 1;
				}
			}
			j = 0;
			k = 0;
		}
	}

	return 0;
}

/*
 * sys_access_auth()
 *	basic user authentication
 * Return:
 *	0 = authentication failed, 1 = authentication successfull, -1 = error
 */
int
sys_access_auth(const char *clientuser, const char *clientpw)
{
	int		i, r, fd, found;
	char		ch[1], *password;
	char		serveruserpw[1024], serveruser[1024], serverpw[1024];
	unsigned char	salt[2];

	i = r = found = 0;

	memset(serveruserpw, 0, sizeof(serveruserpw));

	/* open htpasswd */
	if ((fd = open(config.htpasswd, O_RDONLY)) == -1)
		return -1;

	/* search user in htpasswd */
	while (read(fd, ch, 1)) {
		if (ch[0] != '\n') {
			serveruserpw[i] = ch[0];
			i++;
		} else {
			strcuts(serveruser, serveruserpw, '\0', ':',
			    sizeof(serveruser));
			strcuts(serverpw, serveruserpw, ':', '\0',
			    sizeof(serverpw));
			if (strcmp(serveruser, clientuser) == 0) {
				found = 1;
				break;
			}
			memset(serveruserpw, 0, sizeof(serveruserpw));
			i = 0;
		}
	}

	/* user not found in htpasswd */
	if (found == 0) {
		close(fd);
		return r;
	}

	/* get DES password salt from htpasswd */
	salt[0] = serverpw[0];
	salt[1] = serverpw[1];

	/* generate DES encrypted password */
	password = crypt(clientpw, salt);

	/* compare DES encrypted passwords */
	if (strcmp(password, serverpw) == 0)
		r = 1;

	/* close htpasswd */
	close(fd);

	return r;
}

/*
 * sys_date()
 *	converts struct tm to a CLF conform date string
 * Return:
 *	pointer to date string = sucessfull, NULL = error
 */
char *
sys_date(struct tm *t, int gmt)
{
	int		hour, min, sec, year, mon, mday, diff;
	char		*d, sign;
	static char	date[64];

	hour = t->tm_hour;
	min = t->tm_min;
	sec = t->tm_sec;
	year = t->tm_year;
	mon = t->tm_mon;
	mday = t->tm_mday;

	if (hour == 0)
		diff = 24 - gmt;
	else
		diff = hour - gmt;

	if (diff < 0) {
		diff = abs(diff);
		sign = '-';
	} else {
		sign = '+';
	}

	snprintf(date, sizeof(date), "[%02d/%s/%d:%02d:%02d:%02d %c%02d00]",
	    mday, month[mon], year + 1900, hour, min, sec, sign, diff);

	d = date;

	return d;
}
